(* $Id: godi_tty.ml 1071 2012-07-16 01:22:53Z gerd $  
 * ----------------------------------------------------------------------
 * Copyright 2008 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* We provide two implementations: One assumes xterm-compatible terminal
   capabilities, and one the Windows console.
 *)

open Printf


type key =
  [ `Return
  | `BS
  | `ESC
  | `PageUp
  | `PageDown
  | `Beginning
  | `End
  | `Char of char
  | `None
  ]

let ctrl_re = Str.regexp "[^\x20-\x7e\xa0-\xff]+"


module Xterm = struct
  let out_init() =
    print_string "\027[?1h\027="

  let out_string s =
    print_string
      (Str.global_replace ctrl_re "" s)

  let out_home() =
    print_string "\027[H"

  let out_clear() =
    print_string "\027[H\027[2J"

  let out_clear_eol() =
    print_string "\027[K"

  let out_to_bol() =
    print_string "\013"

  let out_to_bol_next() =
    print_string "\013\n"

  let lines() =
    try int_of_string(Sys.getenv "LINES")
    with Not_found -> 24

  let columns() =
    try int_of_string(Sys.getenv "COLUMNS")
    with Not_found -> 80

  let keys =
    [ ("\027OF", `End);
      ("\027OH", `Beginning);
      ("\027OA", `PageUp);
      ("\027OB", `PageDown);
      ("\027[5~", `PageUp);
      ("\027[6~", `PageDown);
      ("\027OM", `Return); 
      ("\027[3~", `BS);
      ("\127", `BS);
      ("\008", `BS);
      ("\013", `Return);
      ("\n", `Return)
    ]

  let find_key seq =
    List.fold_left
      (fun acc (def_seq, key) ->
	 if acc = `Not_found then
	   if (String.length seq < String.length def_seq) &&
	     (String.sub def_seq 0 (String.length seq) = seq) then
	       `Prefix
	   else
	     if seq = def_seq then
	       `Key key
	     else
	       acc
	 else
	   acc
      )
      `Not_found
      keys

  (* We don't have access to Cygwin's terminal driver, so simply invoke
     stty to set the terminal characteristics
   *)

  let ws_re = Str.regexp "[ \t\r\n]+"

  let get_tty() =
    (* This works, but the corresponding Godi_exec code does not... *)
    let ch = Unix.open_process_in "tty" in
    let tty = input_line ch in
    ignore(Unix.close_process_in ch);
    if tty = "" then
      failwith "Cannot determine tty"
    else
      Str.global_replace ws_re "" tty
(*
    try Sys.getenv "TTY"
    with Not_found -> 
      prerr_endline "The environment variable TTY is not set. This port of";
      prerr_endline "godi_console requires this variable. Set it by doing";
      prerr_endline "  export TTY=`tty`";
      prerr_endline "and start godi_console again. Also note that you need";
      prerr_endline "a terminal that interprets ANSI codes like rxvt.";
      prerr_endline "The standard Windows console will not work!";
      failwith "Cannot enter interactive mode"
 *)

  let tty_lz = lazy(get_tty())


  let set_terminal() =
    let tty = Lazy.force tty_lz in
    let code =
      Sys.command
	(sprintf "%s -F %s -ixon -icanon -echo min 1 time 0"
	   (Godi_file_helper.backslashify
	      (Godi_file.nativename
		 (Godi_file.filename "/bin/stty")))
	   tty
	) in
    if code <> 0 then
      failwith "Cannot set terminal characteristics"


  let reset_terminal() =
    let tty = Lazy.force tty_lz in
    let code =
      Sys.command
	(sprintf "%s -F %s sane"
	   (Godi_file_helper.backslashify
	      (Godi_file.nativename
		 (Godi_file.filename "/bin/stty")))
	   tty
	) in
    if code <> 0 then
      failwith "Cannot set terminal characteristics"


  let poll_stdin wait s =
    let p = Godi_win32.peek_pipe Unix.stdin in
    if not wait && p=0 then
      0
    else
      try (Unix.read Unix.stdin s 0) 1 
      with Unix.Unix_error((Unix.EAGAIN|Unix.EWOULDBLOCK), _, _) -> 0 


  (* TODO: For performance reasons, we should call set_terminal only once,
     and reset it not before the program exits
   *)

  let poll_keyboard_tty ?(wait=false) () =
    try
      let key = ref `None in
      let s = String.create 1 in
      let n = poll_stdin wait s in
      if n > 0 then (
	(* Special handling of ESC: *)
	if s.[0] = '\027' then (
	  (* try to interpret an ESC sequence *)
	  let seq = ref (String.copy s) in
	  let state = ref `Prefix in
	  while !state = `Prefix do
	    let k = ref 0 in
	    let cont = ref true in
	    while !cont do
	      Godi_win32.sleep 0.01;
	      let n = poll_stdin false s in
	      incr k;
	      if n = 0 && !k >= 10 then
		( state := `Not_found; cont := false )
	      else
		if n = 1 then
		  cont := false
	    done;
	    if !state = `Prefix then (
	      seq := !seq ^ s;
	      state := find_key !seq
	    )
	  done;
	  match !state with
	      `Key k -> key := k
	    | `Not_found -> key := `ESC
	    | _ -> ()
	)
	else (
	  (* look the key up: *)
	  try
	    key := List.assoc s keys
	  with
	      Not_found ->
		key := `Char s.[0];
	);
	!key
      ) else (
	`None
      )
    with
      | error ->
	  raise error
end


module Console = struct
  let out_init() =
    ()

  let out_string s =
    let attr = Netsys_win32.get_console_attr() in
    let s' = Str.global_replace ctrl_re "" s in
    print_string s';
    flush stdout;
    let attr' =
      { attr with
          Netsys_win32.cursor_x = attr.Netsys_win32.cursor_x + String.length s';
      } in
    Netsys_win32.set_console_attr attr'

  let out_home() =
    let attr = Netsys_win32.get_console_attr() in
    let attr' =
      { attr with
          Netsys_win32.cursor_x = 0;
          cursor_y = 0
      } in
    Netsys_win32.set_console_attr attr'

  let out_clear() =
    Netsys_win32.clear_console()

  let out_clear_eol() =
    Netsys_win32.clear_until_end_of_line()

  let out_to_bol() =
    let attr = Netsys_win32.get_console_attr() in
    let attr' =
      { attr with
          Netsys_win32.cursor_x = 0;
      } in
    Netsys_win32.set_console_attr attr'


  let out_to_bol_next() =
    let attr = Netsys_win32.get_console_attr() in
    let attr' =
      { attr with
          Netsys_win32.cursor_x = 0;
          cursor_y = attr.Netsys_win32.cursor_y + 1
      } in
    Netsys_win32.set_console_attr attr'

  let lines() =
    (Netsys_win32.get_console_info()).Netsys_win32.height

  let columns() =
    (Netsys_win32.get_console_info()).Netsys_win32.width

  let set_terminal() =
    let m =
      { Netsys_win32.enable_echo_input = false;
        enable_insert_mode = false;
        enable_line_input = false;
        enable_processed_input = false;
        enable_quick_edit_mode = false;
        enable_processed_output = false;
        enable_wrap_at_eol_output = false
      } in
    Netsys_win32.set_console_mode m

  let reset_terminal() =
    let m =
      { Netsys_win32.enable_echo_input = true;
        enable_insert_mode = false;
        enable_line_input = true;
        enable_processed_input = true;
        enable_quick_edit_mode = false;
        enable_processed_output = true;
        enable_wrap_at_eol_output = true
      } in
    Netsys_win32.set_console_mode m

  let poll_stdin wait =
    let p = Godi_win32.have_console_event Unix.stdin in
    if not wait && not p then
      None
    else (
      try
        Some(Godi_win32.read_console_key Unix.stdin)
      with
        | Not_found -> None
    )

  let poll_keyboard_tty ?(wait=false) () =
    let key = ref `None in
    match poll_stdin wait with
      | None ->
          `None
      | Some ev ->
          if ev.Godi_win32.key_down then (
            match ev.Godi_win32.virtual_key_code with
              | 0x08 -> `BS
              | 0x0d -> `Return
              | 0x1b -> `ESC
              | 0x21 -> `PageUp
              | 0x22 -> `PageDown
              | 0x23 -> `End
              | 0x24 -> `Beginning
              | 0x2e -> `BS (* del *)
              | _ -> 
                  if ev.Godi_win32.ascii_char <> 0 then
                    let c = Char.chr ev.Godi_win32.ascii_char in
                    `Char c
                  else
                    `None
          )
          else
            `None
end



let emu = ref `Xterm

let () =
  if Netsys_win32.is_console Unix.stdin then
    emu := `Console

let out_init() =
  match !emu with
    | `Xterm -> Xterm.out_init()
    | `Console -> Console.out_init()

let out_string s =
  match !emu with
    | `Xterm -> Xterm.out_string s
    | `Console -> Console.out_string s

let out_home() =
  match !emu with
    | `Xterm -> Xterm.out_home()
    | `Console -> Console.out_home()

let out_clear() =
  match !emu with
    | `Xterm -> Xterm.out_clear()
    | `Console -> Console.out_clear()

let out_clear_eol() =
  match !emu with
    | `Xterm -> Xterm.out_clear_eol()
    | `Console -> Console.out_clear_eol()

let out_to_bol() =
  match !emu with
    | `Xterm -> Xterm.out_to_bol()
    | `Console -> Console.out_to_bol()

let out_to_bol_next() =
  match !emu with
    | `Xterm -> Xterm.out_to_bol_next()
    | `Console -> Console.out_to_bol_next()

let lines() =
  match !emu with
    | `Xterm -> Xterm.lines()
    | `Console -> Console.lines()

let columns() =
  match !emu with
    | `Xterm -> Xterm.columns()
    | `Console -> Console.columns()


let script = ref ( [] : key list ) ;;

let poll_keyboard ?wait () =
  match !script with
      key :: script' ->
	script := script';
	key
    | [] ->
	match !emu with
	  | `Xterm -> Xterm.poll_keyboard_tty ?wait ()
	  | `Console -> Console.poll_keyboard_tty ?wait ()
;;


let load_script filename =
  let f = Godi_file.open_in `Text filename in
  let s = ref [] in
  try
    while true do
      let line = input_line f in
      match line with
	  "RETURN"    -> s := `Return :: !s
	| "BS"        -> s := `BS :: !s
	| "ESC"       -> s := `ESC :: !s
	| "PGUP"      -> s := `PageUp :: !s
	| "PGDOWN"    -> s := `PageDown :: !s
	| "BEGINNING" -> s := `Beginning :: !s
	| "END"       -> s := `End :: !s
	| _           -> for k = 0 to String.length line - 1 do
	                   s := (`Char line.[k]) :: !s
			 done
    done;
    assert false
  with
      End_of_file ->
	script := List.rev !s;
	close_in f
    | error ->
	close_in f;
	raise error
;;



let forced_refresh_event() = 0.0

let enter_interactive_mode() =
  match !emu with
    | `Xterm -> Xterm.set_terminal()
    | `Console -> Console.set_terminal()

let leave_interactive_mode() =
  match !emu with
    | `Xterm -> Xterm.reset_terminal()
    | `Console -> Console.reset_terminal()

